/* serverpacket.c
 *
 * Construct and send DHCP server packets
 *
 * Russ Dill <Russ.Dill@asu.edu> July 2001
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <time.h>

#include "common.h"
#include "serverpacket.h"
#include "dhcpd.h"
#include "options.h"
#include "static_leases.h"

#include "csman.h"
#include "unilog.h"
#include "csid/csid_dhcp.h"
#include "csid/csid_sdmz.h"
#include "csid/csid_local.h"
#include "csid/csid_3g.h"
/* send a packet to giaddr using the kernel ip stack */
static int send_packet_to_relay(struct dhcpMessage *payload)
{
	DEBUG(LOG_INFO, "Forwarding packet to relay");

	return kernel_packet(payload, server_config.server, SERVER_PORT,
			payload->giaddr, SERVER_PORT);
}


/* send a packet to a specific arp address and ip address by creating our own ip packet */
static int send_packet_to_client(struct dhcpMessage *payload, int force_broadcast)
{
	uint8_t *chaddr;
	uint32_t ciaddr;

	if (force_broadcast) {
		DEBUG(LOG_INFO, "broadcasting packet to client (NAK)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else if (payload->ciaddr) {
		DEBUG(LOG_INFO, "unicasting packet to client ciaddr");
		ciaddr = payload->ciaddr;
		chaddr = payload->chaddr;
	} else if (ntohs(payload->flags) & BROADCAST_FLAG) {
		DEBUG(LOG_INFO, "broadcasting packet to client (requested)");
		ciaddr = INADDR_BROADCAST;
		chaddr = MAC_BCAST_ADDR;
	} else {
		DEBUG(LOG_INFO, "unicasting packet to client yiaddr");
		ciaddr = payload->yiaddr;
		chaddr = payload->chaddr;
	}
	return raw_packet(payload, server_config.server, SERVER_PORT,
			ciaddr, CLIENT_PORT, chaddr, server_config.ifindex);
}


/* send a dhcp packet, if force broadcast is set, the packet will be broadcast to the client */
static int send_packet(struct dhcpMessage *payload, int force_broadcast)
{
	int ret;

	if (payload->giaddr)
		ret = send_packet_to_relay(payload);
	else ret = send_packet_to_client(payload, force_broadcast);
	return ret;
}


static void init_packet(struct dhcpMessage *packet, struct dhcpMessage *oldpacket, char type)
{
	init_header(packet, type);
	packet->xid = oldpacket->xid;
	memcpy(packet->chaddr, oldpacket->chaddr, 16);
	packet->flags = oldpacket->flags;
	packet->giaddr = oldpacket->giaddr;
	packet->ciaddr = oldpacket->ciaddr;
	add_simple_option(packet->options, DHCP_SERVER_ID, server_config.server);
}


/* add in the bootp options */
static void add_bootp_options(struct dhcpMessage *packet)
{
	packet->siaddr = server_config.siaddr;
	if (server_config.sname)
		strncpy((char*)packet->sname, server_config.sname, sizeof(packet->sname) - 1);
	if (server_config.boot_file)
		strncpy((char*)packet->file, server_config.boot_file, sizeof(packet->file) - 1);
}


/* send a DHCP OFFER to a DHCP DISCOVER */
int sendOffer(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct dhcpOfferedAddr *lease = NULL;
	uint32_t req_align, lease_time_align = server_config.lease;
	uint8_t *req, *lease_time;
	struct option_set *curr;
	struct in_addr addr;
	char *hostname=NULL, *ptr;
	unsigned char hostnameLen;

	uint32_t static_lease_ip;

	init_packet(&packet, oldpacket, DHCPOFFER);

	static_lease_ip = getIpByMac(server_config.static_leases, oldpacket->chaddr);
	
	int csman_fd = -1;      
	
//#if defined(AMIT_ENABLE_SDMZ)	
	int sdmz_enable = -1;
  int hb_enable = 0;
  static int hb_chk1=0, hb_chk2=0;
	csman_fd = open_csman(NULL, 0);
  if(csman_fd < 0){
	  LPRINTF(PRI_D, "Open CSMAN Fail, Disable sdmz_enable\n");
	  sdmz_enable = 0;
  }
  else{
		read_csman(csman_fd,CSID_C_SDMZ_ENABLE, &sdmz_enable, sizeof(sdmz_enable), CSM_R_ZERO);     
		read_csman(csman_fd,CSID_C_LOCAL_HALFBRIDGING_ENABLE, &hb_enable, sizeof(hb_enable), CSM_R_ZERO);     
		close_csman(csman_fd);
  }	
//#endif

	/* ADDME: if static, short circuit */
	if(!static_lease_ip)
	{
	/* the client is in our lease/offered table */
	if ((lease = find_lease_by_chaddr(oldpacket->chaddr))) {
        if( ntohl(lease->yiaddr) < ntohl(server_config.start) ||
            ntohl(lease->yiaddr) > ntohl(server_config.end)    ){
            /* Request IP is Not in the Range of IP Pool, choose one for it */
            LOG(LOG_WARNING, "Request IP is Not in the Range of IP Pool, choose one for it");            	
#if defined(AMIT_ENABLE_SDMZ)
		//add by jack 2007/12/07
			if(sdmz_enable)
			{
				if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
				{
					LOG(LOG_DEBUG, "sendOffer:3.packet.yiaddr = server_config.sdmz_ipn\n");				
					packet.yiaddr = server_config.sdmz_ip;
				}
				else
					packet.yiaddr = find_address(0);
			}else
#endif	            
            packet.yiaddr = find_address(0);
            //LOG(LOG_WARNING, "======Jay find_address(0)_1======");

			/* try for an expired lease */
			if (!packet.yiaddr) 
			{				
				packet.yiaddr = find_address(1);
				//LOG(LOG_WARNING, "======Jay find_address(1)_1======");	
			}	
        }
        else{
            /* Original Design */
            if (!lease_expired(lease))
                lease_time_align = lease->expires - time(0);
#if defined(AMIT_ENABLE_SDMZ)
			if(sdmz_enable)
			{
				//add by jack 2007/12/07
				if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
				{
					LOG(LOG_DEBUG, "sendOffer:1.packet.yiaddr = server_config.sdmz_ipn\n");					
					packet.yiaddr = server_config.sdmz_ip;
				}
				//end by jack
				else
					packet.yiaddr = lease->yiaddr; /* FIXME: oh my, is there a host using this IP? */
			}else
#endif				
            packet.yiaddr = lease->yiaddr;
        }

	/* Or the client has a requested ip */
	} else if ((req = get_option(oldpacket, DHCP_REQUESTED_IP)) &&

		   /* Don't look here (ugly hackish thing to do) */
		   memcpy(&req_align, req, 4) &&

		   /* and the ip is in the lease range */
		   ntohl(req_align) >= ntohl(server_config.start) &&
		   ntohl(req_align) <= ntohl(server_config.end) &&

			/* exclude Router Lan IP  2009/3/17 Ken  */
			ntohl(req_align) != ntohl(server_config.server) &&

			!static_lease_ip &&  /* Check that its not a static lease */
			/* and is not already taken/offered */
		   ((!(lease = find_lease_by_yiaddr(req_align)) ||

		   /* or its taken, but expired */ /* ADDME: or maybe in here */
		   lease_expired(lease)))) {
#if defined(AMIT_ENABLE_SDMZ)
			if(sdmz_enable)
			{
				//add by jack 2007/12/07
				if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
				{
					LOG(LOG_DEBUG, "sendOffer:1.packet.yiaddr = server_config.sdmz_ipn\n");					
					packet.yiaddr = server_config.sdmz_ip;
				}
				//end by jack
				else
					packet.yiaddr = req_align; /* FIXME: oh my, is there a host using this IP? */
			}else
#endif
					packet.yiaddr = req_align; /* FIXME: oh my, is there a host using this IP? */

			/* otherwise, find a free IP */
	} else {
			/* Is it a static lease? (No, because find_address skips static lease) */
#if defined(AMIT_ENABLE_SDMZ)
		//add by jack 2007/12/07
		if(sdmz_enable)
		{		
			if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
			{
				//LOG(LOG_DEBUG, "sendOffer:2.packet.yiaddr = server_config.sdmz_ipn\n");				
				packet.yiaddr = server_config.sdmz_ip;
			}
			//end by jack
			else
			{
				packet.yiaddr = find_address(0);

				/* try for an expired lease */
				if (!packet.yiaddr) packet.yiaddr = find_address(1);
			}	
		}else
#endif
		{
			packet.yiaddr = find_address(0);
			//LOG(LOG_WARNING, "======Jay find_address(0)_2======");
			hb_chk1=1;

			/* try for an expired lease */
			if (!packet.yiaddr) 
			{	
				packet.yiaddr = find_address(1);
				hb_chk1=2;
				//LOG(LOG_WARNING, "======Jay find_address(1)_2======");
			}	
		}
	}
	char dhcp_lease[128];
	FILE *fp = fopen("/var/run/udhcpd.leases", "r");
	if(fp != NULL) 
	{
		memset(dhcp_lease, 0, sizeof(dhcp_lease));
		if(fgets(dhcp_lease, sizeof(dhcp_lease), fp) == NULL) 
		{
			LOG(LOG_WARNING, "udhcpd.leases file is NULL!");
			hb_chk2=1;
	  }	
  }
  else
  {
  	hb_chk2=2;
  }	
  	
	 

	if(!packet.yiaddr) {
		LOG(LOG_WARNING, "no IP addresses to give -- OFFER abandoned");
	  if ((hb_enable == 1) && (hb_chk1 == 2) && (hb_chk2 != 0))
	  {
	    csman_fd = -1;
	    unsigned long csid_val = 0;
	    csman_fd = open_csman(NULL, 0);
	    
	    if(csman_fd < 0)
	    {	    	
        LPRINTF(PRI_D, "Open CSMAN Fail! \n");        
	    }
	    else
	    {
	    	csid_val = 3;
        write_csman(csman_fd, CSID_S_3G_CNT_STATUS, &csid_val, 4, CSM_R_ZERO);  
        close_csman(csman_fd);
	    }	  	
	  	LOG(LOG_WARNING, "helf bridge restart 3G connect!");	  	
	  	hb_chk1=0;
	  	hb_chk2=0;
	  }	
		return -1;
	}

	if((ptr=get_option(oldpacket, DHCP_HOST_NAME))!=NULL){
		hostnameLen=*(unsigned char *)(ptr-1);
		if( hostnameLen <= MAXHOSTNAMELEN ){
			if( (hostname=malloc(hostnameLen+1))!=NULL){
				memcpy(hostname, ptr, hostnameLen);
				hostname[hostnameLen]=0;
			}	
		}
	}	

	if (!add_lease(packet.chaddr, packet.yiaddr, hostname,server_config.offer_time)) {
		LOG(LOG_WARNING, "lease pool is full -- OFFER abandoned");
		return -1;
	}

	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
	}

	/* Make sure we aren't just using the lease time from the previous offer */
	if (lease_time_align < server_config.min_lease)
		lease_time_align = server_config.lease;
	}
	/* ADDME: end of short circuit */
	else
	{
		/* It is a static lease... use it */
		packet.yiaddr = static_lease_ip;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
		{
#if defined(AMIT_ENABLE_SDMZ)
			if(sdmz_enable)
			{	
				//add by jack 2007/12/07
				if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
				{
					if(curr->data[OPT_CODE] == DHCP_SUBNET)
					{
						add_simple_option(packet.options, DHCP_SUBNET, server_config.sdmz_nm);
					}
					else if(curr->data[OPT_CODE] == DHCP_SERVER_ID)
					{
						add_simple_option(packet.options, DHCP_SERVER_ID, server_config.sdmz_gw);
					}
					else if(curr->data[OPT_CODE] == DHCP_ROUTER)
					{
						add_simple_option(packet.options, DHCP_ROUTER, server_config.sdmz_gw);
					}
					else if(curr->data[OPT_CODE] == DHCP_DNS_SERVER)
					{
						add_simple_option(packet.options, DHCP_DNS_SERVER, server_config.sdmz_dns1);
					}
					else
						add_option_string(packet.options, curr->data);
						
				}
				//end by jack
				else
					add_option_string(packet.options, curr->data);				
			}
			else
#endif
				add_option_string(packet.options, curr->data);
		}
		curr = curr->next;
	}

	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
//	LOG(LOG_INFO, "sending OFFER of %s", inet_ntoa(addr));
//to broadcast the DHCP REPLY message,because the br0 interface can't send message correctly at kernel 2.4.18 when we
//switch the interface from wlan0 to eth0

	//mike.chiu  Broadcast_Offer_and_ACK_message_after_releasing+{ 
	/*If there is no ciaddr, must sendOffer by broadcast*/
    
    #if 0
        // LF: Using CSID to determine whether to Broadcast or not
	    if(oldpacket->ciaddr==0)		
		    packet.flags |= htons(BROADCAST_FLAG);
	#endif
    //mike.chiu  Broadcast_Offer_and_ACK_message_after_releasing+}

#ifdef _AMIT_BROADCAST_DHCP_REPLY
	return send_packet(&packet, 1);
#else
    csman_fd = -1;
    int broadcast = 0;
    csman_fd = open_csman(NULL, 0);
    
    if(csman_fd < 0){
        LPRINTF(PRI_D, "Open CSMAN Fail, Disable Broadcast\n");
        broadcast = 0;
    }
    else{
        read_csman(csman_fd, CSID_C_DHCP_BROADCAST_ENABLE, &broadcast, 4, CSM_R_ZERO);  
        close_csman(csman_fd);
    }

    if(broadcast == 1){
		packet.flags |= htons(BROADCAST_FLAG);
        return send_packet(&packet, 1);
    }
    else{
	    return send_packet(&packet, 0);
    }
#endif
}


int sendNAK(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;

	init_packet(&packet, oldpacket, DHCPNAK);

	DEBUG(LOG_INFO, "sending NAK");
	return send_packet(&packet, 1);
}


int sendACK(struct dhcpMessage *oldpacket, uint32_t yiaddr)
{
	struct dhcpMessage packet;
	struct option_set *curr;
	uint8_t *lease_time;
	uint32_t lease_time_align = server_config.lease;
	struct in_addr addr;
	char *hostname=NULL, *ptr;
	unsigned char hostnameLen;

	init_packet(&packet, oldpacket, DHCPACK);
	packet.yiaddr = yiaddr;
	int csman_fd = -1;
		
//#if defined(AMIT_ENABLE_SDMZ)
	int sdmz_enable = -1;
	csman_fd = open_csman(NULL, 0);      
	
  if(csman_fd < 0){
      LPRINTF(PRI_D, "Open CSMAN Fail, Disable sdmz_enable\n");
      sdmz_enable = 0;
  }
  else{
	read_csman(csman_fd,CSID_C_SDMZ_ENABLE, &sdmz_enable, sizeof(sdmz_enable), CSM_R_ZERO);     
	close_csman(csman_fd);
  }		
//#endif

	if ((lease_time = get_option(oldpacket, DHCP_LEASE_TIME))) {
		memcpy(&lease_time_align, lease_time, 4);
		lease_time_align = ntohl(lease_time_align);
		if (lease_time_align > server_config.lease)
			lease_time_align = server_config.lease;
		else if (lease_time_align < server_config.min_lease)
			lease_time_align = server_config.lease;
	}

	add_simple_option(packet.options, DHCP_LEASE_TIME, htonl(lease_time_align));

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME) {
#if defined(AMIT_ENABLE_SDMZ)
			if(sdmz_enable)
			{	
				//add by jack 2007/12/07
				if (!memcmp(packet.chaddr, sdmz_chaddr, 16))
				{
					if(curr->data[OPT_CODE] == DHCP_SUBNET)
					{
						add_simple_option(packet.options, DHCP_SUBNET, server_config.sdmz_nm);
					}
					else if(curr->data[OPT_CODE] == DHCP_SERVER_ID)
					{
						add_simple_option(packet.options, DHCP_SERVER_ID, server_config.sdmz_gw);
					}
					else if(curr->data[OPT_CODE] == DHCP_ROUTER)
					{
						add_simple_option(packet.options, DHCP_ROUTER, server_config.sdmz_gw);
					}
					else if(curr->data[OPT_CODE] == DHCP_DNS_SERVER)
					{
						add_simple_option(packet.options, DHCP_DNS_SERVER, server_config.sdmz_dns1);
					}
					else
						add_option_string(packet.options, curr->data);
				}
				//end by jack
				else
					add_option_string(packet.options, curr->data);
			}
			else
#endif
				add_option_string(packet.options, curr->data);
		}
		curr = curr->next;
	}

	add_bootp_options(&packet);

	addr.s_addr = packet.yiaddr;
//	LOG(LOG_INFO, "sending ACK to %s", inet_ntoa(addr));

//to broadcast the DHCP REPLY message,because the br0 interface can't send message correctly at kernel 2.4.18 when we
//switch the interface from wlan0 to eth0

	//mike.chiu  Broadcast_Offer_and_ACK_message_after_releasing+{ 
	/*If there is no ciaddr, must sendACK by broadcast*/

    #if 0   
        //LF: Using CSID to determine whether to Broadcast or not 
	    if(oldpacket->ciaddr==0)		
		    packet.flags |= htons(BROADCAST_FLAG);
    #endif
	//mike.chiu  Broadcast_Offer_and_ACK_message_after_releasing+}

#ifdef _AMIT_BROADCAST_DHCP_REPLY
	if (send_packet(&packet, 1) < 0)
		return -1;        
#else
    csman_fd = -1;
    int broadcast = 0;
    csman_fd = open_csman(NULL, 0);
    
    if(csman_fd < 0){
        LPRINTF(PRI_D, "Open CSMAN Fail! Disable Broadcast");
        broadcast = 0;
    }
    else{
        read_csman(csman_fd, CSID_C_DHCP_BROADCAST_ENABLE, &broadcast, 4, CSM_R_ZERO);
        close_csman(csman_fd);    
    }

    if(broadcast == 1){
		packet.flags |= htons(BROADCAST_FLAG);
        if(send_packet(&packet, 1) < 0){
            return -1;
        }
    }
    else{
        if(send_packet(&packet, 0) < 0){
            return -1;
        }
    }
#endif

	if((ptr=get_option(oldpacket, DHCP_HOST_NAME))!=NULL){
		hostnameLen=*(unsigned char *)(ptr-1);
		if( hostnameLen <= MAXHOSTNAMELEN ){
			if( (hostname=malloc(hostnameLen+1))!=NULL){
				memcpy(hostname, ptr, hostnameLen);
				hostname[hostnameLen]=0;
			}	
		}
	}	
 
	add_lease(packet.chaddr, packet.yiaddr, hostname,lease_time_align);

	return 0;
}


int send_inform(struct dhcpMessage *oldpacket)
{
	struct dhcpMessage packet;
	struct option_set *curr;

	init_packet(&packet, oldpacket, DHCPACK);

	curr = server_config.options;
	while (curr) {
		if (curr->data[OPT_CODE] != DHCP_LEASE_TIME)
			add_option_string(packet.options, curr->data);
		curr = curr->next;
	}

	add_bootp_options(&packet);

	return send_packet(&packet, 0);
}
